#!/usr/bin/perl

#    Copyright 2010, 2011, Henning Kopp, Nina Bindel
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

use Getopt::Long;
use application "polytope";
use List::Util qw[max min];

print "Welcome to tropical insect\n";

my ($colors, $forcejavaview, $help, $intersectioncolor, $intersectiontransp, $show, $stable, $surfaces, $transparencies, $verbose);
my $result = GetOptions(
  'colors=s'             => \$colors,
  'forcejavaview'        => \$forcejavaview,
  'help'                 => \$help,
  'intersectioncolor=s'  => \$intersectioncolor,
  'intersectiontransp=s' => \$intersectiontransp,
  'show=s'               => \$show,
  'stable'               => \$stable,
  'surfaces=s'           => \$surfaces,
  'transparencies=s'     => \$transparencies,
  'verbose'              => \$verbose
);

my @surfaces;
if (!$surfaces) {
  $help=1;
}else{
  @surfaces=split(',',$surfaces);
  #Create an array with the surfaces in it (Don't confuse it with the literal $surfaces)
}

if($surfaces=~m|[^xyzt0-9^,\-\+\*]|){
  #if there is any non-sensible character in $surfaces the user perhaps tries to hack our program
  die "Sorry, there is an error concerning your choice of surfaces. Please check, if you have some typo.\n";
}

my @show;
if( $show && $show !~ m/^(\d+(,\d+)*)$/ && $show ne "none" ) { #lazy evaluation ftw
  $help=1;
}elsif($show eq 'none'){
  @show=();
}elsif($show eq ''){  #no input
  @show=(0..$#surfaces);
}elsif($show || $show==0){
  @show=split(',',$show);
}

#Here we test the intersectioncolor
if(!$intersectioncolor) {
  $intersectioncolor="0 0 255"; #If no color specified, use blue
}else{
  if($intersectioncolor =~ m/^(\d+),(\d+),(\d+)$/){
    if(0<=$1 && 0<=$2 && 0<=$3 && 255>=$1 && 255>=$2 && 255>=$3){
      $intersectioncolor="$1 $2 $3";
    }else{
      die "Intersectioncolor \"$intersectioncolor\" is not a valid RGB-value\n";
    }
  }else{
    $help=1;
  }
}

#Here we test $colors
my @colors=();
my @testcolors=split( /\],?\[/,$colors);
@testcolors=("255,0,0", "0,255,0") if ($colors eq '');
push(@testcolors, "0,255,0") if ($#testcolors==0);
foreach my $i (@testcolors) {
  if($i =~ m/^\[?(\d+),(\d+),(\d+)\]?$/){
    if(0<=$1 && 0<=$2 && 0<=$3 && 255>=$1 && 255>=$2 && 255>=$3){
      push(@colors, "$1 $2 $3");
    }else{
      die "Color \"$i\" is not a valid RGB-value\n";
    }
  }else{
    $help=1;
  }
}

#Here we test $intersectiontransp, the intersectiontransparency
if(!$intersectiontransp) {
  $intersectiontransp='0.01';
}else{
  die "Intersectiontransp has to be between 0 and 1\n" unless($intersectiontransp eq '1' || $intersectiontransp eq '0' || $intersectiontransp =~ m/^0\.\d+$/);
}

#Now we have to test $transparencies
my @transparencies=();
unless ($transparencies eq '') {
  my @testtransparencies=split( /\],?\[/,$transparencies);
  foreach my $i (@testtransparencies) {
    if($i =~ m/^\[?(0\.\d+)\]?$/ || $i =~ m/^\[?1\]?/ || $i =~ m/^\[?0\]?$/) {
      push(@transparencies, $1);
    }else{
      die "Transparency \"$i\" is not a valid value\n";
    }
  }
}

if($help) {
  print 'Usage:
  
    polymake --script tropicalinsect.pl <arguments>

    where <arguments> could be:
    --colors=[0,255,255][255,255,255]... to choose some colors for the surfaces.
          you can also separate them by comma, i.e. [0,255,255],[255,255,255]
          If not specified, the first color will be red and the second green.
          Everything else, except the intersection will be random
    --forcejavaview to force the use of javaview
    --intersectioncolor=.. to specify the color for the intersection in RGB-format. Standard is blue.
    --intersectiontransp=.. to specify the transparency of the intersection.
          1 is clear, 0 is opaque. Starndard is 0.01
    --surfaces=.. to define the surfaces. This option is required
    --stable to draw the stable intersection
    --show=none or --show=0,2,.. if you do not want all surfaces to be shown.
          The first surface has the number 0 and "none" means just the intersection will be drawn
    --transparencies=[0.3][0.7].. to specify different transparency values for different surfaces.
          Like with --colors, you can also separate the values by comma, i.e. [0.3],[0.7].
          Default is 0.5.
    --help to see this help
    --verbose if you like debug information

    Examples:
    polymake --script tropicalinsect.pl --surfaces=x+ty^2,xy+z+t --stable --colors=[0,255,255][255,120,255]
    polymake --script tropicalinsect.pl --surfaces=x+y,x+y+z+t --intersectioncolor=255,150,0 --verbose
    polymake --script tropicalinsect.pl --surfaces=x+y+z+t,x+y+z --intersectiontransp=0.3 --forcejavaview

';
  exit;
}

chdir '/tmp' or die "Couldn't switch to directory '/tmp'.\n";
#We need names for our future files
`rm -f tropicalinsect_surface* tropicalinsect_intersection`;

for(my $i=0; $i<=$#surfaces; $i++) {
#Create for each curve a surface-file
  `gfan_tropicalintersection --tplane >tropicalinsect_surface$i <<-EOT\n Q[t,x,y,z]\n{$surfaces[$i]}\n EOT`;
}

#create the intersection
if($stable) {
  print "Computing stable intersection\n" if($verbose);
  `gfan_tropicalintersection --stable >tropicalinsect_gfan_stable <<-EOT\n Q[t,x,y,z]\n{$surfaces}\n EOT`;
  `gfan_tropicalintersection --tplane >tropicalinsect_upperhalfplane <<-EOT\n Q[t,x,y,z]\n{}\n EOT`;
  `gfan_fancommonrefinement -i1 tropicalinsect_gfan_stable -i2 tropicalinsect_upperhalfplane >tropicalinsect_intersection`;
  `rm tropicalinsect_upperhalfplane tropicalinsect_gfan_stable`;
}else{
  `gfan_tropicalintersection --tplane >tropicalinsect_intersection <<-EOT\n Q[t,x,y,z]\n{$surfaces}\n EOT`;
}

############################################################################
############################################################################
################# Some useful subroutines ##################################
############################################################################
############################################################################
#We now declare a subroutine for reading those surface-files we just created
sub readgfan ($) { #takes a name of a surface-file
  my $filename=$_[0];
  my $filehandle;
  #Now we open our inputfile read-only. Perhaps it cannot be openened, e.g. on Linux if we have no rights to read it.
  open $filehandle, "<$filename" or die "File $filename could not be openened!\n";
  print "File \"$filename\" successfully opened\n";
  my $chapter='';
  #a surface-file, which is a gfan-output consists of the chapters "AMBIENT_DIM", "DIM", "LINEALITY_DIM, "RAYS", "N_RAYS", "LINEALITY_SPACE",   "ORTH_LINEALITY_SPACE", "F_VECTOR", "CONES", "MAXIMAL_COMES" and "PURE".
  #The chapter we are currently reading is saved in $chapter
  #we are not interested in all the chapters
  my %tropical=(
    AMBIENT_DIM =>'',
    #this will be a digit used for consistency check of the data
    DIM =>'',
    #this will again be a digit used for consistency check of the data
    LINEALITY_DIM =>'',
    #again a digit
    RAYS =>[],
    #this will become a reference to an array of references to arrays
    #you can think of it as a matrix if it pleases you
    N_RAYS =>'',
    #here we will have a number, again used for integrity checks
    LINEALITY_SPACE => [],
    #here we will have an array of arrays
    MAXIMAL_CONES =>[],
    #this will become a reference to an array of references to arrays
  );
  my $line;
  while($line=<$filehandle>) {
    chomp($line);
    #line by line we read our file
    #chomp removes the trailing newline
    if($line eq "AMBIENT_DIM") {
      $chapter="AMBIENT_DIM";
      print "Reading AMBIENT_DIM\n";
    } elsif ($line eq "DIM") {
      $chapter="DIM";
      print "Reading DIM\n";
    } elsif ($line eq "LINEALITY_DIM") {
      $chapter="LINEALITY_DIM";
      print "Reading LINEALITY_DIM\n";
    } elsif ($line eq "RAYS") {
      $chapter="RAYS";
      print "Reading RAYS\n";
    } elsif ($line eq "N_RAYS") {
      $chapter="N_RAYS";
      print "Reading N_RAYS\n";
    } elsif ($line eq "LINEALITY_SPACE") {
      $chapter="LINEALITY_SPACE";
      print "Reading LINEALITY_SPACE\n";
    } elsif ($line eq "ORTH_LINEALITY_SPACE") {
      $chapter="ORTH_LINEALITY_SPACE";
      print "Reading ORTH_LINEALITY_SPACE\n";
    } elsif ($line eq "F_VECTOR") {
      $chapter="F_VECTOR";
      print "Reading F_VECTOR\n";
    } elsif ($line eq "F_VECTOR_BOUNDED") {
      $chapter="F_VECTOR_BOUNDED";
      print "Reading F_VECTOR_BOUNDED\n";
    } elsif ($line eq "SIMPLICIAL") {
      $chapter="SIMPLICIAL";
      print "Reading SIMPLICIAL\n";
    } elsif ($line eq "PURE") {
      $chapter="PURE";
      print "Reading PURE\n";
    } elsif ($line eq "CONES") {
      $chapter="CONES";
      print "Reading CONES\n";
    } elsif ($line eq "MAXIMAL_CONES") {
      $chapter="MAXIMAL_CONES";
      print "Reading MAXIMAL_CONES\n";
    } elsif ($line eq "MULTIPLICITIES") {
      $chapter="MULTIPLICITIES";
      print "Reading MULTIPLICITIES\n";
    }
    #Here we determine if the chapter changes 
    #and set our chapter variable accordingly
    if($chapter eq "AMBIENT_DIM" && $line=~m/^\d$/) {
      #we expect to get only one digit.
      $tropical{$chapter}=$line;
      print "....$line\n";
    }
    if($chapter eq "DIM" && $line=~m/^\d$/) {
      #we expect to get only one digit.
      $tropical{$chapter}=$line;
      print "....$line\n";
    }
    if($chapter eq "LINEALITY_DIM" && $line=~m/^\d$/) {
      $tropical{$chapter}=$line;
      print "....$line\n";
    }
    if($chapter eq "RAYS" && $line=~m/^(-?\d+) (-?\d+) (-?\d+) (-?\d+)/) {
      #we expect four numbers, perhaps with a preceding -
      push(@{$tropical{$chapter}}, [$1, $2, $3, $4]);
      print "....$1, $2, $3, $4\n";
      #this is some tricky stuff
      #we push the reference to the array with the current ray onto
      #the array referenced by $tropical{RAYS}
    }
    if($chapter eq "N_RAYS" && $line=~m/^\d+$/) {
      #here we can get a number
      $tropical{$chapter}=$line;
      print "....$line\n";
    }
    if($chapter eq "LINEALITY_SPACE" && $line=~m/^(-?\d+) (-?\d+) (-?\d+) (-?\d+)/) {
      #we expect four numbers, perhaps with a preceding -
      push(@{$tropical{$chapter}}, [$1, $2, $3, $4]);
      print "....$1, $2, $3, $4\n";
    }
    if($chapter eq "MAXIMAL_CONES" && $line=~m/((\d+)( \d+)*)/) {
      my @maxcones=split(/ /, $1);
      push(@{$tropical{$chapter}}, \@maxcones);
      print "....". join (",",@maxcones) ."\n";
    }
  }
  close($filehandle) or die "File \"$filename\" could not be closed.\n";
  print "File \"$filename\" successfully closed.\n";
  return %tropical;
}
############################################################################
#We now build a function for checking the received data
sub datacheck ($) { #takes a reference to some tropicalhash
  my %tropical=%{$_[0]};
  die "Ambient dimension is unequal to four. Your data cannot be represented\n" if ($tropical{'AMBIENT_DIM'} ne 4);

  die "N_RAYS is not equal to the number of rays. Perhaps there was an error reading the file.\n" if ($tropical{'N_RAYS'}-1 ne $#{$tropical{'RAYS'}});
  #Secondly, we check the number of rays

  die "LINEALITY_DIM is not equal to the number of rays in LINEALITY_SPACE. Perhaps there was an error reading the file.\n" if ($tropical{'LINEALITY_DIM'}-1 ne $#{$tropical{'LINEALITY_SPACE'}});
  #then the lineality_dimension
}
############################################################################
#in gfan, if we have a point, the first coordinate i.e. the t-coordinate is -1, since we cut at t=-1
#in polymake, it is 1, since polymake cuts at t=1
#but in both, rays have t=0
#so we can switch from 'gfan-coordinates' to 'polymake-coordinates'
#by multipling the t-coordinate with -1
#the problem is, that (-1)*(-1)="+1" which polymake interprets as a string
sub gfantopolymake ($) { #takes a reference to a tropicalhash
  my %tropical=%{$_[0]};
  map {
    $tropical{'RAYS'}->[$_]->[0] *= -1;
    if ($tropical{'RAYS'}->[$_]->[0]=~m/\++(\d+)/) {
      $tropical{'RAYS'}->[$_]->[0] = $1;
    }
  } 0..$tropical{'N_RAYS'}-1;  #magic, do not touch
}
############################################################################
#This gives us the Lineality space, i.e. some vectors, which span the lineality space.
sub linspace($) { #it takes a reference to a tropicalobject
  my %tropical=%{$_[0]};
  my @veclist=(); #list of vectors we will return
  print "Computing Lineality space spanning vectors:\n" if($verbose);
  #we will take each vector (pos) and its negative vector (neg)
  foreach my $pos (@{$tropical{'LINEALITY_SPACE'}}) {
    push(@veclist, $pos);
    print "....@{$pos}\n" if($verbose);
    #now we mustn't change $pos, since it is a reference and will change our veclist
    my $neg;
    foreach (0..$tropical{'AMBIENT_DIM'}-1) {
      $neg->[$_]=$pos->[$_] * -1;  #autovivification!
      if ($neg->[$_]=~m/\++(\d+)/) {
        $neg->[$_] = $1;
      }
    }
    push(@veclist, $neg);
    print "....@{$neg}\n" if($verbose);
    #we push the vector and the negative vector on the list
  }
  return @veclist;
}
############################################################################
#Now something for Visualizing the tropicalobjects,
#we give the subroutine a tropicalobject (and a color) and it gives us an array
#with the corresponding VISUAL-objects
#due to a bug in the time of writing this program, where jreality couldn't handle transparency, there is the option
#not to give a transparency value
sub visuarr{
  my ($tropical,$color,$transparency,$bbox)=@_; #$bbox=bounding box
  my %tropical=%{$tropical};
  my @polys=();
  #here, we push in all polytopes
  my $temp;
  #this is the temporary polytope we are working with
  my @linealsp=&linspace($tropical);
  foreach my $cone (@{$tropical{'MAXIMAL_CONES'}}){
    my @points=map {$tropical{'RAYS'}->[@{$cone}[$_]]}  (0..$#{$cone});
    push(@points, @linealsp);
    $temp=new Polytope<Rational>(POINTS=>[@points]);
    $temp=intersection($temp, $bbox);
    if($verbose) {
      print "Polytope has points:\n";
      foreach my $point (@points) {
        print "..";
        print " ". $point->[$_] foreach (0.. $#{$point}+1); #print coordinates
        print "\n";
      }
    }
    unless( grep {$_!=0} map {$_->[0]} (@points) ){  #if all first coordinates are zero
      print "....These are points at infinity which will not be visible\n" if($verbose);
    }else{
      if($#points==0) {
        push (@polys, $temp->VISUAL(PointColor=>$color, PointThickness=>'0.25'));
      }elsif($#points==1 || ($#points==2 && $#linealsp==1) ) {
        push (@polys, $temp->VISUAL(EdgeColor=>$color, EdgeThickness=>'3', PointThickness=>'0.25'));
      }else{
        push (@polys, $temp->VISUAL(EdgeColor=>$color, EdgeThickness=>'0.5', FacetColor=>$color, FacetTransparency=>$transparency, PointThickness=>'0.25'));
      }
    }
  }
  return @polys;
}
############################################################################
############################################################################
################End of those useful subroutines ############################
############################################################################
############################################################################
my (@objects, @tropicalvisuals);

foreach (@show) {  #go through all surfaces, which we want to see
  #process surfaces
  my %object=&readgfan("tropicalinsect_surface$_");
  &datacheck(\%object);
  &gfantopolymake(\%object);
  push(@objects, \%object);
}

#The same for the intersection
my %object=&readgfan('tropicalinsect_intersection');
&datacheck(\%object);
&gfantopolymake(\%object);
push(@objects, \%object);

#We do not need those files anymore
`rm -f tropicalinsect_surface* tropicalinsect_intersection`;

#here comes the bounded_box algorithm
#first, we determine the maximal vector

#$i->{'RAYS'}->[$ray]->[$x]
my @maxvars;
foreach my $i (@objects) {
  foreach my $ray (0..$#{$i->{'RAYS'}}) {
    #@{$i->{'RAYS'}->[$ray]} is the array, from which we want to have the max
    push(@maxvars, max (map {abs($_)} @{$i->{'RAYS'}->[$ray]}) );
  }
}
my $max=6/5*(10+max @maxvars); #note that max can't be 0, because there must be at least one point (homogeneous coordinates!!) ; 6/5 is arbitrary
print "Length of the bounding cube is ".($max*2)."\n" if ($verbose);
my $cube=cube(3,$max,-$max);
#the cube is the bounding box

while($#objects>0) { #We want to compute visuals, except of the last element, which is the intersection
  my $color=shift(@colors);
  $color=rand().' '.rand().' '.rand() if ($color eq '');
  my $transparency=shift(@transparencies);
  $transparency='0.5' if ($transparency eq '');
  my $object=shift(@objects);
  push(@tropicalvisuals, &visuarr($object, $color, $transparency, $cube));
}
push(@tropicalvisuals, &visuarr($objects[0], $intersectioncolor,$intersectiontransp,$cube)); #intersection

if($forcejavaview) {
  print "Using Javaview\n" if($verbose);
  javaview(compose(@tropicalvisuals)); #Draw!
}else{
  print "Using Jreality\n" if($verbose);
  jreality(compose(@tropicalvisuals));
}
